//                                                                                
//    Copyright (c) 2006 Thomas Fuchs / The Rockin'-B, www.rockin-b.de       
//                                                                                
//    Permission is granted to anyone to use this software for any purpose        
//    and to redistribute it freely, subject to the following restrictions:       
//                                                                                
//    1.  The origin of this software must not be misrepresented. You must not    
//        claim that you wrote the original software. If you use this software    
//        in a product, an acknowledgment in the product documentation would be   
//        appreciated but is not required.                                        
//                                                                                
//    2.  ANY COMMERCIAL USE IS PROHIBITED. This includes that you must not use   
//        this software in the production of a commercial product. Only an explicit
//        permission by the author can allow you to do that.                      
//                                                                                
//    3.  Any modification applied to the software must be marked as such clearly.
//                                                                                
//    4.  This license may not be removed or altered from any distribution.       
//                                                                                
//                                                                                
//    			    NO WARRANTY                                                 
//                                                                                
//    THERE IS NO WARRANTY FOR THE PROGRAM.                                       
//    THE PROGRAM IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,               
//    EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,                 
//    THE IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR PURPOSE.                 
//    THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
//    SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY    
//    SERVICING, REPAIR OR CORRECTION.                                            
//                                                                                
//    IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES,         
//    INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
//    OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED   
//    TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY    
//    YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER  
//    PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE       
//    POSSIBILITY OF SUCH DAMAGES.                                                
//                                                                                


#include <SGL.H>
#include <SEGA_TIM.H>
#include <RB_CartRAM.h>
#include <RB_were_here.h>

// WARNING: writing to sound RAM can make Yabause and real Saturn crash
//		    I experienced this in both, this tool and the snes emu

typedef struct BENCH {
	char	name[7];
	void	*adr;
	Uint32	u;	// transfer unit (number of bytes per transfer)
	Uint32	n;	// number of transfers
	Uint32	o;	// address offset between transfers
	Uint32 (*function)(struct BENCH *);	// benchmark function
} Bench;

#define READ	tmp = *(adr);		// optimized away?
#define READS	tmp += *(adr);
#define WRITE	*(adr) = tmp;

#define READF	tmp = *(adr++);		// optimized away?
#define READFS	tmp += *(adr++);
#define WRITEF	*(adr++) = tmp;

#define MAKE_FUNCTION(NAME, TYPE, ACCESS)			\
Uint32 NAME(Bench *bench)		\
{\
	volatile register Uint32 i, n = bench->n, o = bench->o;\
	volatile register TYPE *adr = (TYPE *)bench->adr, tmp = (TYPE)0;\
/*were_here("adr %x, n %i, o %i", adr, n, o);*/\
	TIM_FRT_SET_16(0);\
\
	/* perform access */ \
	for(i = 0; i < n; i++, adr += o) {\
		ACCESS;\
	}\
	return TIM_FRT_GET_16();\
}
//((double)/*TIM_FRT_CNT_TO_MCR*/(cycles)) / ((double)(n * sizeof(TYPE)));

// address increment fixed to 1
// why the fucking shit is this slower
#define MAKE_FUNCTION_FAST(NAME, TYPE, ACCESS)			\
Uint32 NAME(Bench *bench)		\
{\
	volatile register TYPE *adr = (TYPE *)bench->adr, *adr2 = adr+bench->n, tmp = (TYPE)0;\
\
	TIM_FRT_SET_16(0);\
	/* perform access */ \
	while(adr+9 < adr2) {\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
		ACCESS;\
	}\
	while(adr < adr2) {\
		ACCESS;\
	}\
	return TIM_FRT_GET_16();\
}

MAKE_FUNCTION(read8, Uint8, READ);
MAKE_FUNCTION(reads8, Uint8, READS);
MAKE_FUNCTION(read16, Uint16, READ);
MAKE_FUNCTION(reads16, Uint16, READS);
MAKE_FUNCTION(read32, Uint32, READ);
MAKE_FUNCTION(reads32, Uint32, READS);
MAKE_FUNCTION(write8, Uint8, WRITE);
MAKE_FUNCTION(write16, Uint16, WRITE);
MAKE_FUNCTION(write32, Uint32, WRITE);

MAKE_FUNCTION_FAST(readf8, Uint8, READF);
MAKE_FUNCTION_FAST(readfs8, Uint8, READFS);
MAKE_FUNCTION_FAST(readf16, Uint16, READF);
MAKE_FUNCTION_FAST(readfs16, Uint16, READFS);
MAKE_FUNCTION_FAST(readf32, Uint32, READF);
MAKE_FUNCTION_FAST(readfs32, Uint32, READFS);
MAKE_FUNCTION_FAST(writef8, Uint8, WRITEF);
MAKE_FUNCTION_FAST(writef16, Uint16, WRITEF);
MAKE_FUNCTION_FAST(writef32, Uint32, WRITEF);

Uint32 buffer[0x80000/sizeof(Uint32)];

#define ADR_HIGH_WORK_RAM		((void *)buffer)
#define ADR_LOW_WORK_RAM		((void *)0x0200000)
#define ADR_VDP1_RAM			((void *)0x5C00000)
#define ADR_VDP2_RAM			((void *)0x5E00000)
#define ADR_SOUND_RAM			((void *)0x5A00000)
#define ADR_CART_RAM			((void *)0x2400000)

#define SIZE_HIGH_WORK_RAM		((Uint32)sizeof(buffer))
#define SIZE_LOW_WORK_RAM		((Uint32)0x80000)//((Uint32)0x100000)
#define SIZE_VDP1_RAM			((Uint32)0x80000)
#define SIZE_VDP2_RAM			((Uint32)0x60000)		// don't erase font
#define SIZE_SOUND_RAM			((Uint32)0x80000)
#define SIZE_CART_RAM			((Uint32)0x80000)		// only test 512 kB of DRAM0, common to 1MB and 4MB carts

Bench benches[] = {
	{"r08_hi", ADR_HIGH_WORK_RAM, sizeof(Uint8), SIZE_HIGH_WORK_RAM/sizeof(Uint8), 1, reads8},	
	{"r08_lo", ADR_LOW_WORK_RAM, sizeof(Uint8), SIZE_LOW_WORK_RAM/sizeof(Uint8), 1, reads8},	
	{"r08_v1", ADR_VDP1_RAM, sizeof(Uint8), SIZE_VDP1_RAM/sizeof(Uint8), 1, reads8},	
	{"r08_v2", ADR_VDP2_RAM, sizeof(Uint8), SIZE_VDP2_RAM/sizeof(Uint8), 1, reads8},	
	{"r08_so", ADR_SOUND_RAM, sizeof(Uint8), SIZE_SOUND_RAM/sizeof(Uint8), 1, reads8},	
	{"r08_ca", ADR_CART_RAM, sizeof(Uint8), SIZE_CART_RAM/sizeof(Uint8), 1, reads8},	

	{"r08Fhi", ADR_HIGH_WORK_RAM, sizeof(Uint8), SIZE_HIGH_WORK_RAM/sizeof(Uint8), 1, readfs8},	
	{"r08Flo", ADR_LOW_WORK_RAM, sizeof(Uint8), SIZE_LOW_WORK_RAM/sizeof(Uint8), 1, readfs8},	
	{"r08Fv1", ADR_VDP1_RAM, sizeof(Uint8), SIZE_VDP1_RAM/sizeof(Uint8), 1, readfs8},	
	{"r08Fv2", ADR_VDP2_RAM, sizeof(Uint8), SIZE_VDP2_RAM/sizeof(Uint8), 1, readfs8},	
	{"r08Fso", ADR_SOUND_RAM, sizeof(Uint8), SIZE_SOUND_RAM/sizeof(Uint8), 1, readfs8},	
	{"r08Fca", ADR_CART_RAM, sizeof(Uint8), SIZE_CART_RAM/sizeof(Uint8), 1, readfs8},	

	{"r16_hi", ADR_HIGH_WORK_RAM, sizeof(Uint16), SIZE_HIGH_WORK_RAM/sizeof(Uint16), 1, read16},	
	{"r16_lo", ADR_LOW_WORK_RAM, sizeof(Uint16), SIZE_LOW_WORK_RAM/sizeof(Uint16), 1, read16},	
	{"r16_v1", ADR_VDP1_RAM, sizeof(Uint16), SIZE_VDP1_RAM/sizeof(Uint16), 1, read16},	
	{"r16_v2", ADR_VDP2_RAM, sizeof(Uint16), SIZE_VDP2_RAM/sizeof(Uint16), 1, read16},	
	{"r16_so", ADR_SOUND_RAM, sizeof(Uint16), SIZE_SOUND_RAM/sizeof(Uint16), 1, read16},	
	{"r16_ca", ADR_CART_RAM, sizeof(Uint16), SIZE_CART_RAM/sizeof(Uint16), 1, read16},	

	{"r32_hi", ADR_HIGH_WORK_RAM, sizeof(Uint32), SIZE_HIGH_WORK_RAM/sizeof(Uint32), 1, read32},	
	{"r32_lo", ADR_LOW_WORK_RAM, sizeof(Uint32), SIZE_LOW_WORK_RAM/sizeof(Uint32), 1, read32},	
	{"r32_v1", ADR_VDP1_RAM, sizeof(Uint32), SIZE_VDP1_RAM/sizeof(Uint32), 1, read32},	
	{"r32_v2", ADR_VDP2_RAM, sizeof(Uint32), SIZE_VDP2_RAM/sizeof(Uint32 ), 1, read32},	
	{"r32_so", ADR_SOUND_RAM, sizeof(Uint32), SIZE_SOUND_RAM/sizeof(Uint32), 1, read32},	
	{"r32_ca", ADR_CART_RAM, sizeof(Uint32), SIZE_CART_RAM/sizeof(Uint32), 1, read32},	
//-----------
	{"w08_hi", ADR_HIGH_WORK_RAM, sizeof(Uint8), SIZE_HIGH_WORK_RAM/sizeof(Uint8), 1, write8},	
	{"w08_lo", ADR_LOW_WORK_RAM, sizeof(Uint8), SIZE_LOW_WORK_RAM/sizeof(Uint8), 1, write8},	
	{"w08_v1", ADR_VDP1_RAM, sizeof(Uint8), SIZE_VDP1_RAM/sizeof(Uint8), 1, write8},	
	{"w08_v2", ADR_VDP2_RAM, sizeof(Uint8), SIZE_VDP2_RAM/sizeof(Uint8), 1, write8},	
//	{"w08_so", ADR_SOUND_RAM, sizeof(Uint8), SIZE_SOUND_RAM/sizeof(Uint8), 1, write8},	
	{"w08_ca", ADR_CART_RAM, sizeof(Uint8), SIZE_CART_RAM/sizeof(Uint8), 1, write8},	

	{"w16_hi", ADR_HIGH_WORK_RAM, sizeof(Uint16), SIZE_HIGH_WORK_RAM/sizeof(Uint16), 1, write16},	
	{"w16_lo", ADR_LOW_WORK_RAM, sizeof(Uint16), SIZE_LOW_WORK_RAM/sizeof(Uint16), 1, write16},	
	{"w16_v1", ADR_VDP1_RAM, sizeof(Uint16), SIZE_VDP1_RAM/sizeof(Uint16), 1, write16},	
	{"w16_v2", ADR_VDP2_RAM, sizeof(Uint16), SIZE_VDP2_RAM/sizeof(Uint16), 1, write16},	
//	{"w16_so", ADR_SOUND_RAM, sizeof(Uint16), SIZE_SOUND_RAM/sizeof(Uint16), 1, write16},	
	{"w16_ca", ADR_CART_RAM, sizeof(Uint16), SIZE_CART_RAM/sizeof(Uint16), 1, write16},	

	{"w32_hi", ADR_HIGH_WORK_RAM, sizeof(Uint32), SIZE_HIGH_WORK_RAM/sizeof(Uint32), 1, write32},	
	{"w32_lo", ADR_LOW_WORK_RAM, sizeof(Uint32), SIZE_LOW_WORK_RAM/sizeof(Uint32), 1, write32},	
	{"w32_v1", ADR_VDP1_RAM, sizeof(Uint32), SIZE_VDP1_RAM/sizeof(Uint32), 1, write32},	
	{"w32_v2", ADR_VDP2_RAM, sizeof(Uint32), SIZE_VDP2_RAM/sizeof(Uint32 ), 1, write32},	
//	{"w32_so", ADR_SOUND_RAM, sizeof(Uint32), SIZE_SOUND_RAM/sizeof(Uint32), 1, write32},	
	{"w32_ca", ADR_CART_RAM, sizeof(Uint32), SIZE_CART_RAM/sizeof(Uint32), 1, write32},	
//---------- this time cache through
	{"r08*hi", 0x20000000 + ADR_HIGH_WORK_RAM, sizeof(Uint8), SIZE_HIGH_WORK_RAM/sizeof(Uint8), 1, read8},	
	{"r08*lo", 0x20000000 + ADR_LOW_WORK_RAM, sizeof(Uint8), SIZE_LOW_WORK_RAM/sizeof(Uint8), 1, read8},	
	{"r08*v1", 0x20000000 + ADR_VDP1_RAM, sizeof(Uint8), SIZE_VDP1_RAM/sizeof(Uint8), 1, read8},	
	{"r08*v2", 0x20000000 + ADR_VDP2_RAM, sizeof(Uint8), SIZE_VDP2_RAM/sizeof(Uint8), 1, read8},	
	{"r08*so", 0x20000000 + ADR_SOUND_RAM, sizeof(Uint8), SIZE_SOUND_RAM/sizeof(Uint8), 1, read8},	
	{"r08*ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint8), SIZE_CART_RAM/sizeof(Uint8), 1, read8},	

	{"r16*hi", 0x20000000 + ADR_HIGH_WORK_RAM, sizeof(Uint16), SIZE_HIGH_WORK_RAM/sizeof(Uint16), 1, read16},	
	{"r16*lo", 0x20000000 + ADR_LOW_WORK_RAM, sizeof(Uint16), SIZE_LOW_WORK_RAM/sizeof(Uint16), 1, read16},	
	{"r16*v1", 0x20000000 + ADR_VDP1_RAM, sizeof(Uint16), SIZE_VDP1_RAM/sizeof(Uint16), 1, read16},	
	{"r16*v2", 0x20000000 + ADR_VDP2_RAM, sizeof(Uint16), SIZE_VDP2_RAM/sizeof(Uint16), 1, read16},	
	{"r16*so", 0x20000000 + ADR_SOUND_RAM, sizeof(Uint16), SIZE_SOUND_RAM/sizeof(Uint16), 1, read16},	
	{"r16*ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint16), SIZE_CART_RAM/sizeof(Uint16), 1, read16},	

	{"r32*hi", 0x20000000 + ADR_HIGH_WORK_RAM, sizeof(Uint32), SIZE_HIGH_WORK_RAM/sizeof(Uint32), 1, read32},	
	{"r32*lo", 0x20000000 + ADR_LOW_WORK_RAM, sizeof(Uint32), SIZE_LOW_WORK_RAM/sizeof(Uint32), 1, read32},	
	{"r32*v1", 0x20000000 + ADR_VDP1_RAM, sizeof(Uint32), SIZE_VDP1_RAM/sizeof(Uint32), 1, read32},	
	{"r32*v2", 0x20000000 + ADR_VDP2_RAM, sizeof(Uint32), SIZE_VDP2_RAM/sizeof(Uint32 ), 1, read32},	
	{"r32*so", 0x20000000 + ADR_SOUND_RAM, sizeof(Uint32), SIZE_SOUND_RAM/sizeof(Uint32), 1, read32},	
	{"r32*ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint32), SIZE_CART_RAM/sizeof(Uint32), 1, read32},	
//--------
	{"w08*hi", 0x20000000 + ADR_HIGH_WORK_RAM, sizeof(Uint8), SIZE_HIGH_WORK_RAM/sizeof(Uint8), 1, write8},	
	{"w08*lo", 0x20000000 + ADR_LOW_WORK_RAM, sizeof(Uint8), SIZE_LOW_WORK_RAM/sizeof(Uint8), 1, write8},	
	{"w08*v1", 0x20000000 + ADR_VDP1_RAM, sizeof(Uint8), SIZE_VDP1_RAM/sizeof(Uint8), 1, write8},	
	{"w08*v2", 0x20000000 + ADR_VDP2_RAM, sizeof(Uint8), SIZE_VDP2_RAM/sizeof(Uint8), 1, write8},	
//	{"w08*so", 0x20000000 + ADR_SOUND_RAM, sizeof(Uint8), SIZE_SOUND_RAM/sizeof(Uint8), 1, write8},	
	{"w08*ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint8), SIZE_CART_RAM/sizeof(Uint8), 1, write8},	

	{"w16*hi", 0x20000000 + ADR_HIGH_WORK_RAM, sizeof(Uint16), SIZE_HIGH_WORK_RAM/sizeof(Uint16), 1, write16},	
	{"w16*lo", 0x20000000 + ADR_LOW_WORK_RAM, sizeof(Uint16), SIZE_LOW_WORK_RAM/sizeof(Uint16), 1, write16},	
	{"w16*v1", 0x20000000 + ADR_VDP1_RAM, sizeof(Uint16), SIZE_VDP1_RAM/sizeof(Uint16), 1, write16},	
	{"w16*v2", 0x20000000 + ADR_VDP2_RAM, sizeof(Uint16), SIZE_VDP2_RAM/sizeof(Uint16), 1, write16},	
//	{"w16*so", 0x20000000 + ADR_SOUND_RAM, sizeof(Uint16), SIZE_SOUND_RAM/sizeof(Uint16), 1, write16},	
	{"w16*ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint16), SIZE_CART_RAM/sizeof(Uint16), 1, write16},	

	{"w32*hi", 0x20000000 + ADR_HIGH_WORK_RAM, sizeof(Uint32), SIZE_HIGH_WORK_RAM/sizeof(Uint32), 1, write32},	
	{"w32*lo", 0x20000000 + ADR_LOW_WORK_RAM, sizeof(Uint32), SIZE_LOW_WORK_RAM/sizeof(Uint32), 1, write32},	
	{"w32*v1", 0x20000000 + ADR_VDP1_RAM, sizeof(Uint32), SIZE_VDP1_RAM/sizeof(Uint32), 1, write32},	
	{"w32*v2", 0x20000000 + ADR_VDP2_RAM, sizeof(Uint32), SIZE_VDP2_RAM/sizeof(Uint32 ), 1, write32},	
//	{"w32*so", 0x20000000 + ADR_SOUND_RAM, sizeof(Uint32), SIZE_SOUND_RAM/sizeof(Uint32), 1, write32},	
	{"w32*ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint32), SIZE_CART_RAM/sizeof(Uint32), 1, write32},
//--------
	{"r08/ca", ADR_CART_RAM, sizeof(Uint8), 1024*1024, 0, read8},	
	{"r16/ca", ADR_CART_RAM, sizeof(Uint16), 1024*1024, 0, read16},	
	{"r32/ca", ADR_CART_RAM, sizeof(Uint32), 1024*1024, 0, read32},	
	{"r08|ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint8), 1024*1024, 0, read8},	
	{"r16|ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint16), 1024*1024, 0, read16},	
	{"r32|ca", 0x20000000 + ADR_CART_RAM, sizeof(Uint32), 1024*1024, 0, read32},	
};
Uint32 nBenches = sizeof(benches) / sizeof(Bench);

#ifndef HAVE_RB_WERE_HERE_H
void clearText()
{
    int i;
    char emptyLine[41] = "                                        ";
    
    for(i = 0; i < 30; i++) {
        slPrint(emptyLine, slLocate(0, i));
    }
}
#endif

Sint16 CartID;

enum {OTYPE_CYCLES_PLAIN, OTYPE_CYCLES_PERBYTE, OTYPE_MS_PLAIN, OTYPE_MS_PERBYTE, OTYPE_NUM};
char *otypenames[OTYPE_NUM] = {
	"cycles plain         ",
	"cycles per byte      ",
	"milliseconds plain   ",
	"milliseconds per byte",
};
void benchmark(Bench *bench, Uint32 nBench, Uint8 init, Uint8 otype)
{
	Uint32 i, x = 0, y = 1, l = 0, cycles;
	double fl;

	clearText();
	slPrint("Benchmarking...", slLocate(0, 0));
	TIM_FRT_INIT(init);	// 8, 32 or 128
	for(i = 0; i < nBench; i++) {
		// may skip cart RAM?
		if(!((bench[i].adr == ADR_CART_RAM) && (CartID != CARTRAM_ID_8M) && (CartID != CARTRAM_ID_32M))) {
			cycles = bench[i].function(bench + i);
			switch(otype) {
				case OTYPE_CYCLES_PLAIN:
					slPrintHex(cycles, slLocate(x+4, y));
					break;

				case OTYPE_CYCLES_PERBYTE:
					slPrintHex(cycles / ((bench + i)->n * (bench + i)->u), slLocate(x+4, y));
					break;

				case OTYPE_MS_PLAIN:
					fl = TIM_FRT_CNT_TO_MCR(cycles);
					slPrintFX(toFIXED(fl), slLocate(x, y));
					break;

				case OTYPE_MS_PERBYTE:
					fl = TIM_FRT_CNT_TO_MCR(cycles);
					fl /= (bench + i)->n * (bench + i)->u;
					slPrintFX(toFIXED(fl), slLocate(x, y));
					break;
			}
		}
		slPrint(bench[i].name, slLocate(x, y));
		// advance screen position
		if(++l == 28) {
			x += 13;
			y = 1;
			l = 0;
		} else y++;
	}
}

void ss_main(void)
{
	Uint32 cyclepattern = 0xe0000000, cp_add = 0x0e000000, benchset = 0;
	Uint8 init = 8, otype = 0;

    slInitSystem(TV_320x240, (TEXTURE *)NULL, 1);
    CartID = RB_CartRAM_init(0, 0);

    slSynch();
    slSynch();
    slSynch();
    slSynch();
    slSynch();
    slSynch();

	while(1) {
		clearText();
		slPrint("RAM area access benchmark tool", slLocate(5, 2));
		slPrint("Select output type with L/R:", slLocate(2, 5));
		slPrint("Select cycle pattern with X:", slLocate(2, 9));
		slPrint("Select init value with Y:", slLocate(2, 13));
		slPrint("Select benchmark set with Z:", slLocate(2, 17));
		slPrint("Press START to begin!", slLocate(9, 25));
		while(~Smpc_Peripheral[0].data & PER_DGT_ST)
			slSynch();
	    while(1) {
			Uint16 push = ~Smpc_Peripheral[0].push;
			// start benchmarking?
			if(push & PER_DGT_ST)
				break;
			// adjust kind of output presentation
			if(push & PER_DGT_TL) {
				if(otype > 0)
					otype --;
				else otype = OTYPE_NUM-1;
			}
			if(push & PER_DGT_TR) {
				if(otype < OTYPE_NUM-1)
					otype ++;
				else otype = 0;
			}
			// change cycle pattern?
			if(push & PER_DGT_TX) {
				// set/increase cycle pattern, wrap around
				if(cp_add) {
					cyclepattern |= cp_add;
					cp_add >>= 4;
				} else {
					cyclepattern = 0xe0000000;
					cp_add = 0x0e000000;
				}
			    slScrCycleSet(cyclepattern, cyclepattern, 0x044e44ee, 0x044e44ee);
			}
			// change timer init setting?
			if(push & PER_DGT_TY) {
				if(init == 8)
					init = 32;
				else if(init == 32)
					init = 128;
				else init = 8;
			}
			// change benchset
			if(push & PER_DGT_TZ) {
				benchset += 87;
				if(benchset >= nBenches)
					benchset = 0;
			}
			slPrint(otypenames[otype], slLocate(2, 7));
			slDispHex(cyclepattern, slLocate(2, 11));
			slPrintHex(init, slLocate(2, 15));
			slPrintHex(benchset, slLocate(2, 19));
			slSynch();
		}

	benchmark(benches + benchset, nBenches - benchset > 87 ? 87 : nBenches - benchset, init, otype);
//		benchmark(benches, nBenches, init, otype);

		while(Smpc_Peripheral[0].data & PER_DGT_ST)
			slSynch();
	}
}


#if 0
#include <SGL.H>
/*********************************/
/* taken from cinit.c */
/*********************************/
#define		SystemWork		0x060ffc00		/* System Variable Address */
#define		SystemSize		(0x06100000-0x060ffc00)		/* System Variable Size */
/* sl.lnk Ŏw肵.bssZNV̊JnAIV{ */
extern Uint32 _bstart, _bend;
extern Uint32 _lstart, _lend;
 
int main( void )
{
	Uint8	*dst;
	Uint32	i; 

	/* 1.Zero Set .bss Section */
	for( dst = (Uint8 *)&_bstart; dst < (Uint8 *)&_bend; dst++ ) {
		*dst = 0;
	}
	
	/* 2.ROM has data at end of text; copy it. */

	/* 3.SGL System Variable Clear */
	for( dst = (Uint8 *)SystemWork, i = 0;i < SystemSize; i++) {
		*dst = 0;
	}

	/* Application Call */
	ss_main();
	return 1;
}
#endif
